'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var uWebSockets_js = require('uWebSockets.js');
var websocket = require('websocket');

const root = 5381;

const keyToId = (key, id = root) => {
  if (isFinite(key)) {
    key = parseFloat(key);
    const hash2 = (((id % root) + 127) * 11) ^ key;
    id = id * 33 ^ key;
    return (id >>> 0) * 4096 + (hash2 >>> 0)
  }
  let i = key.length;
  let hash2 = (id % root) + 127;
  while (i--) {
    const char = key.charCodeAt(i);
    id = (id * 33) ^ char;
    hash2 = (hash2 * 11) ^ char;
  }
  return (id >>> 0) * 4096 + (hash2 >>> 0)
};

const pathToIds = (path, id = root) => [
  id,
  ...path.map(key => {
    id = keyToId(key, id);
    return id
  })
];

class ExtendableSet extends Set {
  constructor (inherits) {
    super();
    this.extenders = new Set();
    if (inherits) {
      this.inherits = inherits;
      this.inherits.extenders.add(this);
    }
  }

  has (value) {
    return super.has(value) || (this.inherits && this.inherits.has(value))
  }

  deleteFromExtenders (value) {
    this.extenders.forEach(extender => {
      if (!extender.delete(value)) {
        extender.deleteFromExtenders(value);
      }
    });
  }

  add (value) {
    if (!this.inherits || !this.inherits.has(value)) {
      super.add(value);
      if (this.extenders.size) {
        this.deleteFromExtenders(value);
      }
    }
  }

  * [Symbol.iterator] () {
    if (this.inherits) {
      yield * this.inherits;
    }
    yield * super[Symbol.iterator]();
  }

  forEach (cb) {
    if (this.inherits) {
      this.inherits.forEach(cb);
    }
    super.forEach(cb);
  }

  updateInheritance (inherits) {
    if (this.inherits) {
      this.inherits.extenders.delete(this);
    }
    if (inherits) {
      inherits.forEach(value => {
        if (!super.delete(value) && this.extenders.size) {
          this.deleteFromExtenders(value);
        }
      });
      inherits.extenders.add(this);
    }
    this.inherits = inherits;
  }
}

const defaultConfig = { offset: 0 };

const setOffset = (config, stamp) => {
  config.offset = stamp - Date.now() + 1e9;
  config.inProgress = false;
};

const createStamp = config => {
  if (!config) {
    config = defaultConfig;
  }

  if (config.inProgress) {
    config.ms += ++config.count / 999;
  } else {
    config.count = 0;
    config.ms = Date.now() + config.offset - 1e9;
    config.inProgress = true;
    setTimeout(() => {
      config.inProgress = false;
    });
  }

  return config.ms
};

const strings = {};

const addToStrings = (id, str) => {
  if (!strings[id]) {
    strings[id] = str;
  }
};

const getString = id => {
  return strings[id]
};

const addParentSubscription = (branch, parent, child, depth, parentList, stamp) => {
  if (branch.subscriptions[parent]) {
    if (branch.subscriptions[parent].stamp === stamp) {
      return
    }

    if (branch.subscriptions[parent].keys) {
      branch.subscriptions[parent].keys.push([child, depth]);
    } else {
      branch.subscriptions[parent].keys = [[child, depth]];
      parentList.push(parent);
    }
  } else {
    branch.subscriptions[parent] = { keys: [[child, depth]] };
    parentList.push(parent);
  }
};

const subscriptions = (branch, id, stamp, depth) => {
  if (!branch.subscriptions[id]) {
    branch.subscriptions[id] = { stamp };
  } else if (branch.subscriptions[id].stamp === stamp) {
    return
  } else {
    branch.subscriptions[id].stamp = stamp;
  }

  if (branch.subscriptions[id].listeners) {
    for (const listenerId in branch.subscriptions[id].listeners) {
      const options = branch.subscriptions[id].listeners[listenerId];
      options.cb(new Leaf(branch, id), options);
    }
  }

  const parent = branch.leaves[id].parent;
  if (parent) {
    if (!branch.parentSubscriptions[depth]) {
      branch.parentSubscriptions[depth] = [];
    }
    addParentSubscription(branch, parent, id, 1, branch.parentSubscriptions[depth], stamp);
  }
};

const checkOptions = (options, id, depth) => {
  let pass = true;

  if (options.depth) {
    pass &= depth <= options.depth;
  }

  if (options.keys) {
    pass &= options.keys.includes(id);
  } else if (options.excludeKeys) {
    pass &= !options.excludeKeys.includes(id);
  }

  return pass
};

const referenceSubscriptions = (branch, ids, stamp, keys, parentList) => {
  ids.forEach(id => {
    if (!branch.subscriptions[id]) {
      branch.subscriptions[id] = {};
    }
    parentSubscriptions(branch, id, stamp, keys, parentList);
  });
};

const parentSubscriptions = (branch, id, stamp, keys, parentList) => {
  if (branch.subscriptions[id].stamp === stamp) {
    return
  } else {
    branch.subscriptions[id].stamp = stamp;
  }

  if (branch.subscriptions[id].listeners) {
    for (const listenerId in branch.subscriptions[id].listeners) {
      const options = branch.subscriptions[id].listeners[listenerId];
      if (keys.find(key => checkOptions(options, key[0], key[1]))) {
        options.cb(new Leaf(branch, id), options);
      }
    }
  }

  if (branch.rF[id]) {
    referenceSubscriptions(branch, branch.rF[id], stamp, keys, parentList);
  }

  const parent = branch.leaves[id].parent;
  if (parent) {
    const last = parentList.length - 1;
    if (!parentList[last]) {
      parentList[last] = [];
    }
    const depth = Math.min(...keys.map(key => key[1]));
    addParentSubscription(branch, parent, id, depth + 1, parentList[last], stamp);
  }
};

const fireParentSubscriptions = (branch, stamp) => {
  const parentList = branch.parentSubscriptions.splice(0);
  while (parentList.length) {
    const nextLevel = parentList.pop();
    if (nextLevel) {
      if (nextLevel.length && !parentList.length) {
        parentList.push([]);
      }
      nextLevel.forEach(id => {
        const keys = branch.subscriptions[id].keys;
        branch.subscriptions[id].keys = null;
        parentSubscriptions(branch, Number(id), stamp, keys, parentList);
      });
    }
  }

  if (branch.branches.length) {
    branch.branches.forEach(branch => fireParentSubscriptions(branch, stamp));
  }
};

const addSubscriptionToQueue = (branch, id, listenerId) => {
  const sub = branch.subscriptions[id].listeners[listenerId];
  branch.client.queue.s.push(
    [true, id, listenerId, sub.keys, sub.excludeKeys, sub.depth, sub.sort, sub.limit]
  );
};

const sendAllSubscriptions = branch => {
  for (const id in branch.subscriptions) {
    for (const listenerId in branch.subscriptions[id].listeners) {
      addSubscriptionToQueue(branch, id, listenerId);
    }
  }

  drainQueue(branch.client);
};

const sendSetExisting = (type, stamp, leaf) => {
  if (!leaf.branch.client.stopSending && type === 'set') {
    const { val, rT } = leaf.branch.leaves[leaf.id];

    leaf.branch.client.queue.l.push([leaf.id, stamp, val, rT]);
  }
};

const addAllDataListener = branch => {
  if (!branch.listeners.allData) {
    branch.listeners.allData = {};
  }

  branch.listeners.allData.client = sendSetExisting;
};

const removeAllDataListener = branch => {
  delete branch.listeners.allData.client;
};

const removeSubscriptionToQueue = (branch, id, listenerId) => {
  branch.client.queue.s.push([false, id, listenerId]);
};

const drainQueue = client => {
  if (
    client.socket &&
    client.socket.readyState === 1 &&
    (
      client.queue.b ||
      client.queue.s.length ||
      client.queue.l.length ||
      client.queue.e.length
    )
  ) {
    client.socket.send(JSON.stringify(client.queue));
    client.queue = { s: [], l: [], e: [] };
  }
};

const emitOwn = (branch, id, event, val, stamp, depth) => {
  const listeners = branch.listeners;

  if (listeners[id] && listeners[id][event]) {
    for (const listenerId in listeners[id][event]) {
      listeners[id][event][listenerId](val, stamp, new Leaf(branch, id));
    }
  }

  if (listeners.allData) {
    for (const listenerId in listeners.allData) {
      listeners.allData[listenerId](val, stamp, new Leaf(branch, id));
    }
  }

  if (event === 'data' && val !== 'add-key' && val !== 'remove') {
    subscriptions(branch, id, stamp, depth);
  }
};

const emitBranches = (branches, id, event, val, stamp, depth) =>
  branches.forEach(branch => {
    if (
      branch.leaves[id] === null ||
      (
        event === 'data' &&
        Object.prototype.hasOwnProperty.call(branch.leaves, id) &&
        Object.prototype.hasOwnProperty.call(branch.leaves[id], 'val')
      )
    ) {
      return
    }

    emitOwn(branch, id, event, val, stamp, depth);

    if (branch.rF[id]) {
      emitReferences(branch, branch.rF[id], event, val, stamp, depth);
    }

    if (branch.branches.length) {
      emitBranches(branch.branches, id, event, val, stamp, depth);
    }
  });

const emitReferences = (branch, ids, event, val, stamp, depth) => {
  ids.forEach(id => {
    emitOwn(branch, id, event, val, stamp, depth);
    if (branch.rF[id]) {
      emitReferences(branch, branch.rF[id], event, val, stamp, depth);
    }
  });
};

const emit = (branch, id, event, val, stamp, depth, isKeys) => {
  emitOwn(branch, id, event, val, stamp, depth);

  if (branch.rF[id]) {
    emitReferences(branch, branch.rF[id], event, val, stamp, depth);
  }

  if (branch.branches.length && !isKeys) {
    emitBranches(branch.branches, id, event, val, stamp, depth);
  }
};

const dataEvents = [];
const afterEmitEvents = [];

const addDataEvent = (branch, id, val, depth) => {
  dataEvents.push([branch, id, val, depth]);
};

const addAfterEmitEvent = (cb) => afterEmitEvents.push(cb);

const emitDataEvents = (branch, stamp) => {
  const afterEmitEventsToRun = afterEmitEvents.splice(0);
  dataEvents.splice(0).forEach(event =>
    emit(
      event[0] || branch,
      event[1],
      'data',
      event[2],
      stamp,
      event[3],
      event[2] === 'add-key' || event[2] === 'remove-key'
    )
  );
  fireParentSubscriptions(branch, stamp);
  afterEmitEventsToRun.forEach(event => event());
  if (branch.client.queue) {
    drainQueue(branch.client);
  }
};

const respectNullOverrides = (branch, id, parent) => {
  if (branch.leaves[parent] === null) {
    branch.leaves[id] = null;
    if (branch.persist) {
      branch.persist.store(String(id), null);
    }
    return true
  }
};

const overrideBranches = (branches, id, parent, overrideLeaf = true, overrideKeys = true) => {
  branches.forEach(branch => {
    const leaf = branch.leaves[id];
    let overrideLeafNext = overrideLeaf;
    let overrideKeysNext = overrideKeys;

    if (leaf === null || respectNullOverrides(branch, id, parent)) {
      return
    }

    if (Object.prototype.hasOwnProperty.call(branch.leaves, id)) {
      if (overrideLeaf) {
        Object.setPrototypeOf(leaf, branch.inherits.leaves[id]);
        overrideLeafNext = false;
      }

      if (overrideKeys && Object.prototype.hasOwnProperty.call(leaf, 'keys')) {
        leaf.keys.updateInheritance(branch.inherits.leaves[id].keys);
        overrideKeysNext = false;
      }
    }

    if (branch.branches.length) {
      overrideBranches(branch.branches, id, parent, overrideLeafNext, overrideKeysNext);
    }
  });
};

const addOwnLeaf = (branch, id, parent, key, depth, stamp) => {
  branch.leaves[id] = { parent, key, stamp, depth, keys: new ExtendableSet() };

  if (branch.branches.length) {
    overrideBranches(branch.branches, id, parent);
  }

  return branch.leaves[id]
};

const addOverrideLeaf = (branch, id) => {
  const leaf = branch.leaves[id] = Object.create(branch.leaves[id]);

  if (branch.branches.length) {
    overrideBranches(branch.branches, id, leaf.parent, true, false);
  }

  return leaf
};

const addOverrideLeafForKeys = (branch, id) => {
  let leaf = branch.leaves[id];
  let overrideLeaf = false;

  if (!Object.prototype.hasOwnProperty.call(branch.leaves, id)) {
    leaf = branch.leaves[id] = Object.create(leaf);
    leaf.keys = new ExtendableSet(branch.inherits.leaves[id].keys);
    overrideLeaf = true;
  } else if (!Object.prototype.hasOwnProperty.call(leaf, 'keys')) {
    leaf.keys = new ExtendableSet(branch.inherits.leaves[id].keys);
  } else {
    return leaf
  }

  if (branch.branches.length) {
    overrideBranches(branch.branches, id, leaf.parent, overrideLeaf);
  }

  return leaf
};

const addReferenceFrom = (branch, rF, rT) => {
  if (!branch.rF[rT]) {
    branch.rF[rT] = new Set();
  }
  branch.rF[rT].add(rF);
};

const removeReferenceFrom = (branch, rF, rT) =>
  branch.rF[rT].delete(rF);

const fixBranchReferences = (branches, rF, rT, rTold) =>
  branches.forEach(branch => {
    const leaf = branch.leaves[rF];
    if (leaf === null) {
      return
    } else if (
      Object.prototype.hasOwnProperty.call(branch.leaves, rF) &&
      Object.prototype.hasOwnProperty.call(leaf, 'val') &&
      leaf.rT
    ) {
      if (leaf.val === rT) {
        addAfterEmitEvent(() => {
          delete leaf.val;
          delete leaf.rT;
        });
      }
      return
    } else {
      if (rTold) {
        removeReferenceFrom(branch, rF, rTold);
      }
      addReferenceFrom(branch, rF, rT);
    }

    if (branch.branches.length) {
      fixBranchReferences(branch.branches, rF, rT);
    }
  });

const checkReferenceByLeaf = (branch, rTBranch, rT, cb) => {
  if (
    branch.leaves[rT] !== null &&
    (
      rTBranch === branch ||
      Object.prototype.isPrototypeOf.call(rTBranch.leaves, branch.leaves)
    )
  ) {
    return cb()
  } else {
    throw new Error('Reference must be in same branch')
  }
};

const fireBranchKeys = (branches, id, keys, stamp) =>
  branches.forEach(branch => {
    if (branch.leaves[id] !== null) {
      const keysNext = new Set(keys);

      keysNext.forEach(key => {
        if (
          Object.prototype.hasOwnProperty.call(branch.leaves, key) &&
          branch.leaves[key] !== null
        ) {
          keysNext.delete(key);
        }
      });

      if (keysNext.size) {
        addDataEvent(branch, id, 'add-key');
        if (branch.branches.length) {
          fireBranchKeys(branch.branches, id, keysNext);
        }
      }
    }
  });

const getByKey = (branch, id, key, val, stamp, inReference) => {
  const leafId = keyToId(key, id);
  if (branch.leaves[leafId]) {
    return leafId
  } else {
    if (branch.leaves[id].rT) {
      const originId = getByKey(
        branch, branch.leaves[id].val, key, val, stamp, true
      );
      if (originId) {
        return originId
      }
    }

    if (!inReference && val !== undefined) {
      set(branch, id, { [key]: val }, stamp);
      return leafId
    }
  }
};

const setByPath = (branch, ids, path, val, stamp, inReference) => {
  let i = ids.length - 1;
  const leafId = ids[i];
  while (i--) {
    if (i === 0 && inReference) {
      return
    }
    // buble hack
    const newVal = { [path.pop()]: val };
    val = newVal;
    if (branch.leaves[ids[i]]) {
      set(branch, ids[i], val, stamp);
      return leafId
    }
  }
};

const getByPath = (branch, id, path, val, stamp, inReference) => {
  const ids = pathToIds(path, id);
  let i = ids.length - 1;
  if (branch.leaves[ids[i]]) {
    return ids[i]
  } else {
    while (i--) {
      const leaf = branch.leaves[ids[i]];
      if (leaf && leaf.rT) {
        const originId = getByPath(
          branch, leaf.val, path.slice(i), val, stamp, true
        );
        if (originId) {
          return originId
        }
      }
    }
    if (val !== undefined) {
      return setByPath(branch, ids, path, val, stamp, inReference)
    }
  }
};

const getApi = (branch, id, path, val, stamp) => {
  if (Array.isArray(path)) {
    return getByPath(branch, id, path, val, stamp)
  } else {
    return getByKey(branch, id, path, val, stamp)
  }
};

const setOwnNewVal = (_, leaf, id, val, stamp) => {
  leaf.rT = false;
  leaf.val = val;
  leaf.stamp = stamp;

  addDataEvent(undefined, id, 'set', leaf.depth);
};

const setOwnNewReference = (branch, leaf, id, rT, stamp) => {
  if (branch.branches.length) {
    fixBranchReferences(branch.branches, id, rT);
  }

  leaf.rT = true;
  leaf.val = rT;
  leaf.stamp = stamp;
  addReferenceFrom(branch, id, rT);
  addDataEvent(undefined, id, 'set', leaf.depth);
};

const setOwnNewKeys = (branch, leaf, id, val, stamp) => {
  const keys = new Set();
  for (const key in val) {
    const valKey = val[key];
    if (key === 'val') {
      setOwnNew(branch, leaf, id, val.val, stamp);
    } else if (valKey !== undefined && valKey !== null) {
      const subLeafId = keyToId(key, id);
      const keyId = keyToId(key);
      addToStrings(keyId, key);
      keys.add(subLeafId);
      leaf.keys.add(subLeafId);
      const subLeaf = addOwnLeaf(branch, subLeafId, id, keyId, leaf.depth + 1, stamp);
      setOwnNew(branch, subLeaf, subLeafId, valKey, stamp);
    }
  }
  if (keys.size) {
    leaf.stamp = stamp;
    if (branch.branches.length) {
      fireBranchKeys(branch.branches, id, keys);
    }
    addDataEvent(undefined, id, 'add-key');
  }
};

const setOwnNew = (branch, leaf, id, val, stamp) => {
  if (typeof val === 'object') {
    if (Array.isArray(val)) {
      if (val[0] === '@') {
        const rT = getByPath(branch, root, val.slice(1), {}, stamp);
        setOwnNewReference(branch, leaf, id, rT, stamp);
      } else {
        setOwnNewVal(branch, leaf, id, val, stamp);
      }
    } else if (val.isLeaf) {
      checkReferenceByLeaf(branch, val.branch, val.id, () => {
        setOwnNewReference(branch, leaf, id, val.id, stamp);
      });
    } else {
      setOwnNewKeys(branch, leaf, id, val, stamp);
    }
  } else {
    setOwnNewVal(branch, leaf, id, val, stamp);
  }
};

const removeListenersSubscriptions = (branch, id) => {
  delete branch.listeners[id];
  delete branch.subscriptions[id];
};

const removeReferenceFromBranches = (branch, rF, rT) => {
  removeReferenceFrom(branch, rF, rT);

  branch.branches.forEach(branch => {
    if (
      branch.leaves[rF] !== null && (
        !Object.prototype.hasOwnProperty.call(branch.leaves, rF) ||
        !Object.prototype.hasOwnProperty.call(branch.leaves[rF], 'val')
      )
    ) {
      removeReferenceFromBranches(branch, rF, rT);
    }
  });
};

const removeFromBranches = (branches, leaf, id, parent, keys, ownKeys, rT, stamp) =>
  branches.forEach(branch => {
    let leafNext = leaf;
    let parentNext = parent;
    let keysNext = keys;
    let ownKeysNext = ownKeys;
    let rTnext = rT;

    if (branch.leaves[id] === null) {
      delete branch.leaves[id];
      return
    } else {
      let branchLeaf;

      if (Object.prototype.hasOwnProperty.call(branch.leaves, id)) {
        branchLeaf = branch.leaves[id];
        if (leaf) {
          Object.setPrototypeOf(branchLeaf, null);
          branchLeaf.parent = leaf.parent;
          branchLeaf.key = leaf.key;
          branchLeaf.depth = leaf.depth;
          branchLeaf.stamp = stamp;
          if (!branchLeaf.keys) {
            branchLeaf.keys = new ExtendableSet();
          }
          leafNext = false;
        }
        if (ownKeys && Object.prototype.hasOwnProperty.call(branchLeaf, 'keys')) {
          branchLeaf.keys.updateInheritance(null);
          ownKeysNext = false;
        }
      }

      if (keys.length) {
        const addKeys = new Set(keys.filter(
          key => Object.prototype.hasOwnProperty.call(branch.leaves, key) && branch.leaves[key] !== null
        ));
        if (addKeys.size) {
          if (!branchLeaf && leaf) {
            branchLeaf = addOwnLeaf(branch, id, leaf.parent, leaf.key, leaf.depth, leaf.stamp);
            leafNext = false;
            ownKeysNext = false;
          } else {
            branchLeaf = addOverrideLeafForKeys(branch, id);
          }
          addKeys.forEach(key => { branchLeaf.keys.add(key); });
          keysNext = keys.filter(key => !addKeys.has(key));
        }
      }

      if (branchLeaf) {
        if (rT) {
          if (branchLeaf.val === undefined) {
            removeReferenceFrom(branch, id, rT);
          } else {
            rTnext = false;
          }
        }

        if (parent) {
          const pLeaf = addOverrideLeafForKeys(branch, parent);
          pLeaf.keys.add(id);
          parentNext = false;
        }
      } else {
        if (rT) {
          removeReferenceFrom(branch, id, rT);
        }
        if (parent) {
          const pLeaf = branch.leaves[parent];
          if (Object.prototype.hasOwnProperty.call(branch.leaves, parent)) {
            pLeaf.stamp = stamp;
          }
          addDataEvent(branch, parent, 'remove-key', pLeaf.depth);
        }
        removeListenersSubscriptions(branch, id);
      }
    }

    if (branch.branches.length) {
      removeFromBranches(branch.branches, leafNext, id, parentNext, keysNext, ownKeysNext, rTnext, stamp);
    }
  });

const removeChildren = (branch, id, stamp, ownKey) => {
  branch.leaves[id].keys.forEach(key => {
    if (branch.leaves[key] !== null) {
      remove(branch, key, stamp, true, ownKey);
    }
  });
};

const remove = (branch, id, stamp, ignoreParent, ownKey = false) => {
  emit(branch, id, 'data', 'remove', stamp);

  const leaf = branch.leaves[id];
  const rT = leaf.rT && leaf.val;

  if (!ignoreParent && leaf.parent) {
    let pLeaf = branch.leaves[leaf.parent];
    if (!Object.prototype.hasOwnProperty.call(branch.leaves, leaf.parent)) {
      pLeaf = addOverrideLeaf(branch, leaf.parent);
    } else if (Object.prototype.hasOwnProperty.call(pLeaf, 'keys')) {
      ownKey = pLeaf.keys.delete(id);
    }
    pLeaf.stamp = stamp;
    addDataEvent(undefined, leaf.parent, 'remove-key', pLeaf.depth);
  }

  removeChildren(branch, id, stamp, ownKey);

  if (branch.branches.length) {
    removeFromBranches(
      branch.branches, leaf, id, !ignoreParent && leaf.parent, [...leaf.keys], true, rT, stamp
    );
  }

  if (ownKey) {
    delete branch.leaves[id];
  } else {
    branch.leaves[id] = null;
  }

  if (branch.persist) {
    if (ownKey) {
      branch.persist.remove(String(id));
    } else {
      branch.persist.store(String(id), null);
    }
  }

  if (rT) {
    removeReferenceFrom(branch, id, rT);
  }

  removeListenersSubscriptions(branch, id);
};

const setOwnExistingVal = (branch, leaf, id, val, stamp) => {
  if (val === leaf.val) {
    return
  } else if (leaf.rT) {
    removeReferenceFromBranches(branch, id, leaf.val);
  }

  leaf.rT = false;
  leaf.val = val;
  leaf.stamp = stamp;

  addDataEvent(undefined, id, 'set', leaf.depth);
  return true
};

const setOwnExistingReference = (branch, leaf, id, rT, stamp) => {
  const rTold = leaf.rT && leaf.val;
  if (rT === rTold) {
    return
  } else if (rTold) {
    removeReferenceFrom(branch, id, rTold);
  }

  if (branch.branches.length) {
    fixBranchReferences(branch.branches, id, rT, rTold);
  }

  leaf.rT = true;
  leaf.val = rT;
  leaf.stamp = stamp;

  addReferenceFrom(branch, id, rT);
  addDataEvent(undefined, id, 'set', leaf.depth);
  return true
};

const setOwnExisting = (branch, leaf, id, val, stamp) => {
  if (typeof val === 'object') {
    if (!val) {
      remove(branch, id, stamp);
    } else if (Array.isArray(val)) {
      if (val[0] === '@') {
        const rT = getByPath(branch, root, val.slice(1), {}, stamp);
        setOwnExistingReference(branch, leaf, id, rT, stamp);
      } else {
        setOwnExistingVal(branch, leaf, id, val, stamp);
      }
    } else if (val.isLeaf) {
      checkReferenceByLeaf(branch, val.branch, val.id, () =>
        setOwnExistingReference(branch, leaf, id, val.id, stamp));
    } else {
      setKeys(branch, leaf, id, val, stamp, setOwnExisting);
    }
  } else {
    setOwnExistingVal(branch, leaf, id, val, stamp);
  }
};

const setOverrideVal = (branch, leaf, id, val, stamp) => {
  if (val === leaf.val) {
    return
  } else if (leaf.rT) {
    removeReferenceFromBranches(branch, id, leaf.val);
  }

  leaf = addOverrideLeaf(branch, id);
  leaf.rT = false;
  leaf.val = val;
  leaf.stamp = stamp;

  addDataEvent(undefined, id, 'set', leaf.depth);
  return true
};

const setOverrideReference = (branch, leaf, id, rT, stamp) => {
  const rTold = leaf.rT && leaf.val;
  if (rT === rTold) {
    return
  } else if (rTold) {
    removeReferenceFrom(branch, id, rTold);
  }

  if (branch.branches.length) {
    fixBranchReferences(branch.branches, id, rT, rTold);
  }

  leaf = addOverrideLeaf(branch, id);
  leaf.rT = true;
  leaf.val = rT;
  leaf.stamp = stamp;

  addReferenceFrom(branch, id, rT);
  addDataEvent(undefined, id, 'set', leaf.depth);
  return true
};

const setOverride = (branch, leaf, id, val, stamp) => {
  if (typeof val === 'object') {
    if (!val) {
      remove(branch, id, stamp);
    } else if (Array.isArray(val)) {
      if (val[0] === '@') {
        const rT = getByPath(branch, root, val.slice(1), {}, stamp);
        setOverrideReference(branch, leaf, id, rT, stamp);
      } else {
        setOverrideVal(branch, leaf, id, val, stamp);
      }
    } else if (val.isLeaf) {
      checkReferenceByLeaf(branch, val.branch, val.id, () =>
        setOverrideReference(branch, leaf, id, val.id, stamp));
    } else {
      setKeys(branch, leaf, id, val, stamp, setOverride);
    }
  } else {
    setOverrideVal(branch, leaf, id, val, stamp);
  }
};

const setKeys = (branch, leaf, id, val, stamp, set) => {
  const keys = new Set();
  for (const key in val) {
    if (key === 'val') {
      set(branch, leaf, id, val.val, stamp);
    } else {
      const subLeafId = keyToId(key, id);
      if (branch.leaves[subLeafId]) {
        const fn = Object.prototype.hasOwnProperty.call(branch.leaves, subLeafId) ? setOwnExisting : setOverride;
        fn(
          branch, branch.leaves[subLeafId], subLeafId, val[key], stamp
        );
      } else if (val[key] !== undefined && val[key] !== null) {
        const keyId = keyToId(key);
        addToStrings(keyId, key);
        keys.add(subLeafId);
        const subLeaf = addOwnLeaf(
          branch, subLeafId, id, keyId, leaf.depth + 1, stamp
        );
        setOwnNew(branch, subLeaf, subLeafId, val[key], stamp);
      }
    }
  }
  if (keys.size) {
    if (set === setOverride) {
      leaf = addOverrideLeafForKeys(branch, id);
    }
    keys.forEach(key => { leaf.keys.add(key); });
    leaf.stamp = stamp;
    if (branch.branches.length) {
      fireBranchKeys(branch.branches, id, keys);
    }
    addDataEvent(undefined, id, 'add-key');
  }
};

const set = (branch, id, val, stamp) => {
  if (Object.prototype.hasOwnProperty.call(branch.leaves, id)) {
    setOwnExisting(branch, branch.leaves[id], id, val, stamp);
  } else {
    setOverride(branch, branch.leaves[id], id, val, stamp);
  }
};

const bindAllDataListener = (branch, persist) => {
  if (!branch.listeners.allData) {
    branch.listeners.allData = {};
  }

  branch.listeners.allData.persist = (type, _, item) => {
    if (type !== 'remove' && Object.prototype.hasOwnProperty.call(branch.leaves, item.id)) {
      const leaf = branch.leaves[item.id];
      persist.store(
        String(item.id),
        {
          parent: leaf.parent,
          key: leaf.key,
          keyString: getString(leaf.key),
          keys: [...leaf.keys],
          stamp: leaf.stamp,
          depth: leaf.depth,
          val: leaf.val,
          rT: leaf.rT
        }
      );
    }
  };
};

const loadLeaf = (branch, id, leaf) => {
  id = Number(id);

  if (leaf === null) {
    branch.leaves[id] = null;
  } else {
    addToStrings(leaf.key, leaf.keyString);
    delete leaf.keyString;

    if (branch.leaves[id]) {
      if (leaf.keys && leaf.keys.length) {
        addOverrideLeafForKeys(branch, id);
      } else {
        addOverrideLeaf(branch, id);
      }
    } else {
      addOwnLeaf(branch, id, leaf.parent, leaf.key, leaf.depth, leaf.stamp);
    }

    if (leaf.keys) {
      leaf.keys.forEach(key => branch.leaves[id].keys.add(key));
    }

    if (leaf.val !== undefined) {
      const rTold = branch.leaves[id] && branch.leaves[id].rT && branch.leaves[id].val;
      if (rTold) {
        removeReferenceFrom(branch, id, rTold);
      }
      if (leaf.rT) {
        addReferenceFrom(branch, id, leaf.val);
      }
      branch.leaves[id].rT = leaf.rT;
      branch.leaves[id].val = leaf.val;
    }
  }
};

const Leaf = function (branch, id) {
  this.branch = branch;
  this.id = id;
};

const setToNewBranch = (branch, val, stamp) => {
  if (!stamp) {
    stamp = createStamp(branch.stamp);
  }

  set(branch, root, val, stamp);
  emitDataEvents(branch, stamp);
  return new Leaf(branch, root)
};

const prepareNewBranch = inherits => {
  const branch = Object.create(null, {
    branches: { value: [] },
    listeners: { value: {} },
    subscriptions: { value: {} },
    parentSubscriptions: { value: [] },
    rF: { value: {} },
    client: { value: {} }
  });

  if (inherits) {
    Object.defineProperties(branch, {
      leaves: { value: Object.create(inherits.leaves) },
      stamp: { value: inherits.stamp },
      inherits: { value: inherits }
    });
    inherits.branches.push(branch);
    for (const id in inherits.rF) {
      branch.rF[id] = new Set(inherits.rF[id]);
    }
  } else {
    const leaves = Object.create(null);
    leaves[root] = { keys: new ExtendableSet(), depth: 0 };
    Object.defineProperties(branch, {
      leaves: { value: leaves },
      stamp: { value: { offset: 0 } }
    });
  }

  if (inherits && typeof inherits.newBranchMiddleware === 'function') {
    inherits.newBranchMiddleware(new Leaf(branch, root));
  }

  return branch
};

const createPersist = (val, persist, stamp, inherits) => {
  const branch = prepareNewBranch(inherits);

  branch.persist = persist;
  return persist.start()
    .then(() => persist.load((id, leaf) => loadLeaf(branch, id, leaf)))
    .then(() => bindAllDataListener(branch, persist))
    .then(() => setToNewBranch(branch, val, stamp))
};

const create = (val, stamp, inherits) => setToNewBranch(
  prepareNewBranch(inherits), val, stamp
);

function define (object, key, value) {
  Object.defineProperty(object, key, { value });
}

const origin = (branch, id) => {
  return branch.leaves[id].rT ? origin(branch, branch.leaves[id].val) : id
};

const compute = (branch, leaf) => {
  if (leaf.rT) {
    return compute(branch, branch.leaves[leaf.val])
  } else if (leaf.val !== undefined) {
    return leaf.val
  }
};

const forEach = (branch, id, cb) => {
  branch.leaves[id].keys.forEach(key => {
    if (branch.leaves[key] !== null) {
      cb(new Leaf(branch, key), getString(branch.leaves[key].key));
    }
  });
};

const map = (branch, id, cb) => {
  const mapped = [];
  branch.leaves[id].keys.forEach(key => {
    if (branch.leaves[key] !== null) {
      mapped.push(cb(new Leaf(branch, key), getString(branch.leaves[key].key)));
    }
  });
  return mapped
};

const filter = (branch, id, cb) => {
  const filtered = [];
  branch.leaves[id].keys.forEach(key => {
    if (branch.leaves[key] !== null) {
      const subLeafInstance = new Leaf(branch, key);
      if (cb(subLeafInstance, getString(branch.leaves[key].key))) {
        filtered.push(subLeafInstance);
      }
    }
  });
  return filtered
};

const find = (branch, id, cb) => {
  for (const key of branch.leaves[id].keys) {
    if (branch.leaves[key] !== null) {
      const subLeafInstance = new Leaf(branch, key);
      if (cb(subLeafInstance, getString(branch.leaves[key].key))) {
        return subLeafInstance
      }
    }
  }
};

const reduce = (branch, id, cb, accumulator) => {
  let skipFirst = accumulator === undefined;
  branch.leaves[id].keys.forEach(key => {
    const leaf = branch.leaves[key];
    if (leaf !== null) {
      if (skipFirst) {
        accumulator = compute(branch, leaf);
        skipFirst = null;
      } else {
        accumulator = cb(accumulator, new Leaf(branch, key), getString(leaf.key));
      }
    }
  });
  return accumulator
};

const path = (branch, id) => {
  const path = [];
  while (id !== root) {
    path.unshift(getString(branch.leaves[id].key));
    id = branch.leaves[id].parent;
  }
  return path
};

const inspect = (branch, leaf) => {
  if (!leaf) {
    return 'stx { }'
  }

  const start = 'stx ' + (leaf.key ? getString(leaf.key) + ' ' : '');
  const val = leaf.rT ? inspect(branch, branch.leaves[leaf.val]) : leaf.val;
  const subLeaves = [...leaf.keys];
  if (subLeaves.length) {
    let keys = [];
    if (subLeaves.length > 10) {
      const len = subLeaves.length;
      keys = subLeaves.slice(0, 5).map(id => getString(branch.leaves[id].key));
      keys.push(`... ${len - 5} more items`);
    } else {
      keys = subLeaves.map(id => getString(branch.leaves[id].key));
    }
    return val
      ? `${start}{ val: ${val}, ${keys.join(', ')} }`
      : `${start}{ ${keys.join(', ')} }`
  } else {
    return val
      ? `${start}{ val: ${val} }`
      : `${start}{ }`
  }
};

const serialize = (branch, leaf) => {
  if (!leaf) {
    return {}
  }

  let child = false;
  const result = {};
  const val = leaf.rT ? ['@', ...path(branch, leaf.val)] : leaf.val;

  leaf.keys.forEach(key => {
    const subLeaf = branch.leaves[key];
    if (subLeaf !== null) {
      child = true;
      result[getString(subLeaf.key)] = serialize(branch, subLeaf);
    }
  });
  if (child) {
    if (val !== undefined) {
      result.val = val;
    }
    return result
  } else if (val !== undefined) {
    return val
  } else {
    return {}
  }
};

let listenerLastId = 0;

const Listener = function (branch, id, event, listenerId) {
  this.branch = branch;
  this.id = id;
  this.event = event;
  this.listenerId = listenerId;
};

define(Listener.prototype, 'off', function () {
  if (this.event === 'allData') {
    delete this.branch.listeners.allData[this.listenerId];
  } else {
    delete this.branch.listeners[this.id][this.event][this.listenerId];
  }
});

const on = (branch, id, event, cb) => {
  const listenerId = listenerLastId++;

  const listeners = branch.listeners;

  if (event === 'allData') {
    if (!listeners.allData) {
      listeners.allData = {};
    }

    listeners.allData[listenerId] = cb;
  } else {
    if (!listeners[id]) {
      listeners[id] = { [event]: {} };
    } else if (!listeners[id][event]) {
      listeners[id][event] = {};
    }

    listeners[id][event][listenerId] = cb;
  }

  return new Listener(branch, id, event, listenerId)
};

let listenerLastId$1 = 0;

const Subscription = function (branch, id, listenerId) {
  this.branch = branch;
  this.id = id;
  this.listenerId = listenerId;
};

define(Subscription.prototype, 'unsubscribe', function () {
  if (this.branch.subscriptions[this.id]) {
    delete this.branch.subscriptions[this.id].listeners[this.listenerId];

    if (this.branch.client.queue) {
      removeSubscriptionToQueue(this.branch, this.id, this.listenerId);
      drainQueue(this.branch.client);
    }
  }
});

const parseOptions = (id, options, cb) => {
  if (options.excludeKeys) {
    if (options.keys) {
      options.keys = options.keys.filter(key => !options.excludeKeys.includes(key));
      delete options.excludeKeys;
    } else {
      options.excludeKeys = options.excludeKeys.map(key => keyToId(key, id));
    }
  }

  if (options.keys) {
    options.keys = options.keys.map(key => keyToId(key, id));
  } else if (options.sort && options.sort.path) {
    if (![Number, String].includes(options.sort.type)) {
      options.sort.type = String;
    }
    options.sort.type = options.sort.type.name[0];
  }

  options.cb = cb;
};

const subscribe = (branch, id, options, cb) => {
  const listenerId = listenerLastId$1++;

  parseOptions(id, options, cb);
  const subscriptions = branch.subscriptions;

  if (!subscriptions[id]) {
    subscriptions[id] = { listeners: {} };
  } else if (!subscriptions[id].listeners) {
    subscriptions[id].listeners = {};
  }

  subscriptions[id].listeners[listenerId] = options;
  cb(new Leaf(branch, id), options);

  if (branch.client.queue) {
    addSubscriptionToQueue(branch, id, listenerId);
    drainQueue(branch.client);
  }

  return new Subscription(branch, id, listenerId)
};

const uniq = process.pid;
var uid = () => keyToId(`b-${Date.now()}-${(Math.random() * 10000) | 0}-${uniq}`);

var maxSize = 1e7;

const numberSortAsc = (branch, path) => (key1, key2) => {
  const leaf1 = branch.leaves[getByPath(branch, key1, path)];
  const leaf2 = branch.leaves[getByPath(branch, key2, path)];
  return (leaf1 && compute(branch, leaf1)) - (leaf2 && compute(branch, leaf2))
};

const numberSortDesc = (branch, path) => (key1, key2) => {
  const leaf1 = branch.leaves[getByPath(branch, key1, path)];
  const leaf2 = branch.leaves[getByPath(branch, key2, path)];
  return (leaf2 && compute(branch, leaf2)) - (leaf1 && compute(branch, leaf1))
};

const stringSortAsc = (branch, path) => (key1, key2) => {
  const leaf1 = branch.leaves[getByPath(branch, key1, path)];
  const leaf2 = branch.leaves[getByPath(branch, key2, path)];
  return String(leaf1 && compute(branch, leaf1))
    .localeCompare(String(leaf2 && compute(branch, leaf2)))
};

const stringSortDesc = (branch, path) => (key1, key2) => {
  const leaf1 = branch.leaves[getByPath(branch, key1, path)];
  const leaf2 = branch.leaves[getByPath(branch, key2, path)];
  return String(leaf2 && compute(branch, leaf2))
    .localeCompare(String(leaf1 && compute(branch, leaf1)))
};

const cache = (socket, isMaster, id, stamp) => {
  if (isMaster) {
    delete socket.cache.branch[id];
    socket.cache.master[id] = stamp;
  } else {
    delete socket.cache.master[id];
    socket.cache.branch[id] = stamp;
  }
};

const isCachedForStamp = (socket, isMaster, id, stamp) => socket.cache &&
  (isMaster ? socket.cache.master[id] === stamp : socket.cache.branch[id] === stamp);

const isCached = (socket, isMaster, id) => socket.cache &&
  (isMaster ? socket.cache.master[id] : socket.cache.branch[id]);

const reuseCache = (socket) => {
  if (!socket.cache) return undefined

  return {
    cache: {
      master: socket.cache.master,
      branch: {},
      strings: socket.cache.strings
    },
    remove: socket.cache.branch
  }
};

const cacheString = (socket, id) => {
  socket.cache.strings[id] = true;
};

const isStringCached = (socket, id) => socket.cache && socket.cache.strings[id];

const sendToSocket = (socket, payload, next) => {
  socket.send(payload, true);
  setImmediate(next);
};

const sendLarge = (socket, raw, size) => {
  const count = size / maxSize;
  if ((count | 0) === count) {
    raw += ' ';
  }

  if (!socket.blobInProgress) {
    socket.blobInProgress = [];
  }

  console.log('📡 exceeds frame limit - split up', (size / (1024 * 1024)) | 0, 'MB');
  const buf = Buffer.from(raw, 'utf8');
  let i = 0;

  const drainInProgress = done => {
    if (socket.blobInProgress.length > 0) {
      sendToSocket(socket, socket.blobInProgress.shift(), () => drainInProgress(done));
    } else {
      done();
    }
  };

  const next = () => {
    i++;
    if (i * maxSize <= size) {
      sendToSocket(socket, buf.slice(i * maxSize, (i + 1) * maxSize), next);
    } else {
      drainInProgress(() => {
        socket.blobInProgress = null;
      });
    }
  };

  sendToSocket(socket, buf.slice(i * maxSize, (i + 1) * maxSize), next);
};

const send = (socket, raw) => {
  const size = Buffer.byteLength(raw, 'utf8');
  if (size > maxSize) {
    sendLarge(socket, raw, size);
  } else if (socket.blobInProgress) {
    socket.blobInProgress.push(raw);
  } else {
    socket.send(new TextEncoder('utf-8').encode(raw));
  }
};

const sendData = (socket, branch, data) => {
  if (
    Object.keys(data.leaves).length ||
    Object.keys(data.strings).length
  ) {
    const json = { t: createStamp(branch.stamp), l: data.leaves, s: data.strings };
    if (Object.keys(socket.cleanLeaves).length) {
      json.c = socket.cleanLeaves;
      socket.cleanLeaves = {};
    }
    if (Object.keys(socket.removeLeaves).length) {
      json.r = socket.removeLeaves;
      socket.removeLeaves = {};
    }
    send(socket, JSON.stringify(json));
  }
};

const sendLeaves = (socket, leaf, options, dataOverride) => {
  const { branch, id } = leaf;
  if (branch.leaves[id] === null) {
    return
  }

  let { keys, excludeKeys, depth, sort, limit } = options;

  const depthLimit = depth || Infinity;
  const data = dataOverride || { leaves: {}, strings: {}, rTs: new Set(), visited: new Set() };

  serializeParents(data, socket, branch, Number(id));

  let doNotCache;
  if (keys) {
    keys = keys.filter(
      key => serializeWithAllChildren(data, socket, branch, key, depthLimit, 1)
    );
    doNotCache = true;
  } else {
    [keys, doNotCache] = serializeAllChildren(
      data, socket, branch, id, depthLimit, 0, excludeKeys, sort, limit || Infinity
    );
  }

  serializeLeaf(data, socket, branch, id, keys, 0, doNotCache);

  while (data.rTs.size) {
    const rTs = data.rTs;
    data.rTs = new Set();
    rTs.forEach(([rT, depth]) => {
      serializeParents(data, socket, branch, rT);
      serializeWithAllChildren(data, socket, branch, rT, depthLimit, depth);
    });
  }

  data.visited = new Set();

  if (!dataOverride) {
    sendData(socket, branch, data);
  }
};

const serializeAllChildren = (
  data, socket, branch, id, depthLimit, depth, excludeKeys, sort, limit = Infinity
) => {
  const keys = [];
  let originalKeys = branch.leaves[id].keys;
  let doNotCache = false;

  if (sort && sort.path) {
    const sortFunction = sort.type === 'N'
      ? sort.desc ? numberSortDesc : numberSortAsc
      : sort.desc ? stringSortDesc : stringSortAsc;
    originalKeys = [...originalKeys].sort(sortFunction(branch, sort.path));
  }

  for (const key of originalKeys) {
    if (
      branch.leaves[key] === null ||
      (
        excludeKeys &&
        excludeKeys.includes(key)
      ) || (
        !serializeWithAllChildren(data, socket, branch, key, depthLimit, depth + 1)
      )
    ) {
      doNotCache = true;
      continue
    }

    keys.push(key);

    if (!--limit) {
      doNotCache = true;
      break
    }
  }

  return [keys, doNotCache]
};

const serializeWithAllChildren = (data, socket, branch, id, depthLimit, depth) => {
  if (!(id in branch.leaves) || branch.leaves[id] === null || depthLimit < depth) {
    return
  }

  const [keys, doNotCache] = serializeAllChildren(data, socket, branch, id, depthLimit, depth);
  return serializeLeaf(data, socket, branch, id, keys, depth, doNotCache)
};

const serializeParents = (data, socket, branch, id) => {
  let parent = branch.leaves[id].parent;
  while (parent) {
    if (data.leaves[parent]) {
      data.leaves[parent][5].push(id);
      break
    }

    if (!serializeLeaf(data, socket, branch, parent, [id], -1, true)) {
      break
    }

    id = parent;
    parent = branch.leaves[id].parent;
  }
};

const serializeLeaf = (data, socket, branch, id, keys, depth, doNotCache) => {
  const leaf = branch.leaves[id];
  const isMaster = !Object.prototype.hasOwnProperty.call(branch.leaves, id);

  if (leaf !== null && !data.visited.has(id)) {
    if (leaf.rT && depth !== -1 && !data.visited.has(leaf.val)) {
      data.rTs.add([leaf.val, depth]);
    }

    if (!isCachedForStamp(socket, isMaster, id, leaf.stamp)) {
      if (id in data.leaves) {
        keys.push(...data.leaves[id][5]);
      }

      data.leaves[id] = [leaf.key, leaf.parent, leaf.stamp, leaf.val, leaf.rT, keys, leaf.depth];

      if (!doNotCache) {
        cache(socket, isMaster, id, leaf.stamp);
      }

      if (socket.cleanLeaves[id]) {
        delete socket.cleanLeaves[id];
      }

      if (leaf.key !== undefined && !isStringCached(socket, leaf.key)) {
        cacheString(socket, leaf.key);
        data.strings[leaf.key] = getString(leaf.key);
      }

      return true
    }
  }

  data.visited.add(id);
};

const removeLeaves = (socket, type, stamp, leaf) => {
  if (type === 'remove') {
    const { branch, id } = leaf;
    if (isCached(socket, !Object.prototype.hasOwnProperty.call(branch.leaves, id), id)) {
      socket.removeLeaves[id] = stamp;
    }
  }
};

const syncSubscriptions = (branch, socket, subscriptions) => {
  const data = { leaves: {}, strings: {}, rTs: new Set(), visited: new Set() };

  subscriptions.forEach(subscription => {
    const [add, id, listenerId, keys, excludeKeys, depth, sort, limit] = subscription;
    if (add && (id in branch.leaves)) {
      if (!branch.subscriptions[id]) {
        branch.subscriptions[id] = { listeners: {} };
      } else if (!branch.subscriptions[id].listeners) {
        branch.subscriptions[id].listeners = {};
      }

      branch.subscriptions[id].listeners[`${socket.id}-${listenerId}`] = {
        keys,
        excludeKeys,
        depth,
        sort,
        limit,
        cb: sendLeaves.bind(null, socket)
      };

      sendLeaves(
        socket,
        { branch, id },
        { keys, excludeKeys, depth, sort, limit },
        data
      );
    } else if (branch.subscriptions[id] && branch.subscriptions[id].listeners) {
      delete branch.subscriptions[id].listeners[`${socket.id}-${listenerId}`];
    }
  });

  sendData(socket, branch, data);
};

const addAllDataListener$1 = (branch, socket) => {
  if (!branch.listeners.allData) {
    branch.listeners.allData = {};
  }
  branch.listeners.allData[socket.id] = removeLeaves.bind(null, socket);
};

const removeSubscriptionsAndAllDataListener = (branch, socketId) => {
  for (const id in branch.subscriptions) {
    for (const listenerId in branch.subscriptions[id].listeners) {
      if (listenerId.indexOf(socketId) === 0) {
        delete branch.subscriptions[id].listeners[listenerId];
      }
    }
  }

  delete branch.listeners.allData[socketId];
};

const switchBranch = async (socket, master, branchKey, persist) => {
  let branch = master.branches.find(branch => branch.key === branchKey);

  if (!branch) {
    if (persist) {
      branch = await createPersist(undefined, persist, undefined, master);
      branch = branch.branch;
    } else {
      branch = create(undefined, undefined, master).branch;
    }
    branch.key = branchKey;
  }

  removeSubscriptionsAndAllDataListener(socket.branch, socket.id);
  const reuse = reuseCache(socket);
  socket.branch = branch;
  addAllDataListener$1(branch, socket);

  if (reuse) {
    socket.cache = reuse.cache;
    socket.cleanLeaves = reuse.remove;
  }

  return new Leaf(branch, root)
};

const setLeaves = (branch, socket, leaves) => {
  if (branch.clientCanUpdate) {
    const afterEvents = [];
    leaves.forEach(leaf => {
      const [id, stamp, val, rT] = leaf;

      if (!(id in branch.leaves)) {
        return
      }

      const setPath = path(branch, id);
      const rule = branch.clientCanUpdate.find(
        rule => rule.path.length === setPath.length && rule.path.every(
          (key, i) => key === '*' || key === setPath[i]
        )
      );

      if (
        rule &&
        (
          typeof rule.authorize !== 'function' ||
          rule.authorize(new Leaf(branch, id), val)
        )
      ) {
        let changed;
        const leaf = branch.leaves[id];
        if (Object.prototype.hasOwnProperty.call(branch.leaves, id)) {
          cache(socket, false, id, stamp);

          if (rT) {
            changed = setOwnExistingReference(branch, leaf, id, val, stamp);
          } else if (val !== null) {
            changed = setOwnExistingVal(branch, leaf, id, val, stamp);
          }
        } else {
          cache(socket, true, id, stamp);

          if (rT) {
            changed = setOverrideReference(branch, leaf, id, val, stamp);
          } else if (val !== null) {
            changed = setOverrideVal(branch, leaf, id, val, stamp);
          }
        }

        if (changed && typeof rule.after === 'function') {
          afterEvents.push(rule.after.bind(null, new Leaf(branch, id)));
        }
      }
    });
    emitDataEvents(branch, createStamp(branch.stamp));
    afterEvents.forEach(event => event());
  }
};

const fireEmits = (branch, emits) => {
  emits.forEach(([id, event, val, stamp]) => emit(branch, id, event, val, stamp));
};

const processIncoming = async (server, socket, master, data) => {
  const { b: branchKey, s: subscriptions, l: leaves, e: emits } = data;

  if (
    branchKey !== undefined &&
    branchKey !== socket.branch.key &&
    typeof server.switchBranch === 'function'
  ) {
    await server.switchBranch(
      new Leaf(socket.branch, root),
      branchKey,
      switchBranch.bind(null, socket, master)
    );
  }

  if (subscriptions && subscriptions.length) {
    syncSubscriptions(socket.branch, socket, subscriptions);
  }

  if (leaves) {
    setLeaves(socket.branch, socket, leaves);
  }

  if (emits && emits.length) {
    fireEmits(socket.branch, emits);
  }

  if (socket.incoming.length) {
    await processIncoming(server, socket, master, socket.incoming.shift());
  }

  socket.incoming = null;
};

const incoming = (server, socket, master, data) => {
  if (socket.incoming) {
    socket.incoming.push(data);
  } else {
    socket.incoming = [];
    processIncoming(server, socket, master, data);
  }
};

const removeSocket = (server, socket) => {
  removeSubscriptionsAndAllDataListener(socket.branch, socket.id);
  delete server.sockets[socket.id];
};

const Server = function () {
  this.sockets = {};
};

define(Server.prototype, 'close', function (cb) {
  if (this.listen) {
    for (const id in this.sockets) {
      this.sockets[id].end();
    }

    uWebSockets_js.us_listen_socket_close(this.listen);
    this.listen = null;

    if (cb) {
      cb();
    }
  }
});

const listen = (branch, port) => {
  const server = new Server();

  uWebSockets_js.App({

  }).ws('/*', {
    maxPayloadLength: maxSize,
    idleTimeout: 10,
    compression: 1,
    open: (socket) => {
      socket.id = uid();
      socket.branch = branch;
      socket.cache = { master: {}, branch: {}, strings: {} };
      socket.cleanLeaves = {};
      socket.removeLeaves = {};
      socket.incoming = null;
      server.sockets[socket.id] = socket;
      addAllDataListener$1(branch, socket);

      socket.send(new TextEncoder('utf-8').encode(JSON.stringify({
        t: createStamp(socket.branch.stamp)
      })));
    },
    message: (socket, data) => {
      try {
        data = JSON.parse(new TextDecoder('utf-8').decode(data));
        if (!data) return
      } catch (e) {
        return e
      }

      incoming(server, socket, branch, data);
    },
    close: (socket) => server && removeSocket(server, socket)
  }).any('/*', res => {
    res.end('Nothing to see here!');
  }).listen(port, listenSocket => {
    if (listenSocket) {
      console.log(`💫 listening websockets on ${port} 💫`);
      server.listen = listenSocket;
    } else {
      throw Error(`Could not listen on ${port}`)
    }
  });

  return server
};

const concatUint8Arrays = (uInt8ArrayArray) => {
  const totalLength = uInt8ArrayArray.reduce((total, uInt8Array) => total + uInt8Array.length, 0);
  const bigUint8Array = new Uint8Array(totalLength);
  let offset = 0;
  uInt8ArrayArray.forEach(uInt8Array => {
    bigUint8Array.set(uInt8Array, offset);
    offset += uInt8Array.length;
  });
  return bigUint8Array
};

const bindSocketListeners = (socket, close, error, open, message) => {
  socket.uInt8Arrays = [];
  socket.onclose = close;
  socket.onerror = error;
  socket.onopen = open;
  socket.onmessage = ({ data }) => {
    if (typeof data === 'string') {
      message(data);
    } else {
      socket.uInt8Arrays.push(new Uint8Array(data));

      if (data.byteLength < maxSize) {
        message(Buffer.from(
          concatUint8Arrays(socket.uInt8Arrays.splice(0))
        ).toString('utf8'));
      }
    }
  };
};

class WebSocket extends websocket.w3cwebsocket {
  constructor (url) {
    super(url, null, null, null, null, {
      maxReceivedFrameSize: maxSize,
      maxReceivedMessageSize: maxSize
    });
  }
}

const cleanLeaves = (branch, list) => {
  for (let id in list) {
    id = Number(id);
    if (id in branch.leaves) {
      const leaf = branch.leaves[id];
      const rT = leaf.rT && leaf.val;
      const parent = leaf.parent;
      delete branch.leaves[id];
      if (!(parent in list) && parent in branch.leaves) {
        branch.leaves[parent].keys.delete(id);
        branch.leaves[parent].stamp = list[id];
      }
      if (rT) {
        removeReferenceFromBranches(branch, id, rT);
      }
    }
  }
};

const removeLeaves$1 = (branch, list) => {
  for (const id in list) {
    const stamp = list[id];
    if (branch.leaves[id]) {
      remove(branch, Number(id), stamp);
    }
  }
};

const setLeaves$1 = (branch, leaves) => {
  for (let id in leaves) {
    id = Number(id);
    const [key, parent, stamp, val, rT, keys, depth] = leaves[id];

    if (id in branch.leaves) {
      const leaf = branch.leaves[id];

      if (rT) {
        setOwnExistingReference(branch, leaf, id, val, stamp);
      } else if (val !== null) {
        setOwnExistingVal(branch, leaf, id, val, stamp);
      }

      if (keys && keys.length) {
        let added = false;
        keys.forEach(key => {
          if (!leaf.keys.has(key)) {
            leaf.keys.add(key);
            added = true;
          }
        });
        if (added) {
          addDataEvent(undefined, id, 'add-key');
        }
      }
    } else {
      const leaf = addOwnLeaf(branch, id, parent, key, depth, stamp);

      if (rT) {
        setOwnNewReference(branch, leaf, id, val, stamp);
      } else if (val !== null) {
        setOwnNewVal(branch, leaf, id, val, stamp);
      }

      if (keys && keys.length) {
        keys.forEach(key => leaf.keys.add(key));
        addDataEvent(undefined, id, 'add-key');
      }
    }
  }
};

const setStrings = strings => {
  for (const id in strings) {
    addToStrings(id, strings[id]);
  }
};

const incoming$1 = (branch, data) => {
  const { t: stamp, l: leaves, c: clean, s: strings, r: remove } = data;

  if (stamp !== undefined) {
    setOffset(branch.stamp, stamp);
  }

  branch.client.stopSending = true;

  if (clean) {
    cleanLeaves(branch, clean);
  }

  if (remove) {
    removeLeaves$1(branch, remove);
  }

  if (strings) {
    setStrings(strings);
  }

  if (leaves) {
    setLeaves$1(branch, leaves);
  }

  emitDataEvents(branch, stamp);
  branch.client.stopSending = false;
};

const socketClose = WebSocket.prototype.close;

const closeWhenDrained = socket => {
  if (socket.bufferedAmount) {
    setTimeout(closeWhenDrained, 50, socket);
  } else {
    socketClose.call(socket);
  }
};

define(WebSocket.prototype, 'close', function () {
  clearTimeout(this.heartbeat);
  removeAllDataListener(this.branch);
  this.blockReconnect = true;
  closeWhenDrained(this);
});

const heartbeat = branch => {
  const socket = branch.client.socket;

  if (socket) {
    clearTimeout(socket.heartbeat);
    socket.send(JSON.stringify({ h: 1 }));
    socket.heartbeat = setTimeout(() => heartbeat(branch), 8000);
  }
};

const connect = (branch, url, reconnect = 50) => {
  if (branch.client.reconnect) {
    clearTimeout(branch.client.reconnect);
    branch.client.reconnect = null;
  } else if (branch.client.socket) {
    throw Error('Can not connect twice')
  }

  branch.client.socket = {};
  const socket = new WebSocket(url);

  const close = () => {
    if (branch.client.socket) {
      clearTimeout(branch.client.socket.heartbeat);
      branch.client.socket = null;
      emit(branch, root, 'connected', false, createStamp(branch.stamp));
    }

    if (socket.blockReconnect) {
      branch.client.queue = null;
    } else {
      reconnect = Math.min((reconnect * 1.5), 2000);
      branch.client.reconnect = setTimeout(connect, reconnect, branch, url, reconnect);
    }
  };

  const error = () => {
    if (socket.readyState === 1) {
      socket.close();
    }
  };

  const open = () => {
    socket.branch = branch;
    branch.client.socket = socket;
    branch.client.queue = { s: [], l: [], e: [] };
    heartbeat(branch);

    sendAllSubscriptions(branch);
    addAllDataListener(branch);

    emit(branch, root, 'connected', true, createStamp(branch.stamp));
  };

  const message = data => {
    if (data) {
      try {
        data = JSON.parse(data);
        if (!data) return
      } catch (e) {
        return e
      }

      incoming$1(branch, data);
    }
  };

  bindSocketListeners(socket, close, error, open, message);

  return branch.client
};

const defineApi = (leaf) => {
  // ISLEAF
  define(leaf, 'isLeaf', true);

  // CREATE
  define(leaf, 'create', function (val, stamp) {
    if (this.id === root) {
      return create(val, stamp, this.branch)
    } else {
      throw new Error('Can not create from leaf')
    }
  });

  // SET
  define(leaf, 'set', function (val, stamp) {
    if (!stamp) {
      stamp = createStamp(this.branch.stamp);
    }

    const stopSending = this.branch.client.stopSending;
    this.branch.client.stopSending = false;
    set(this.branch, this.id, val, stamp);
    emitDataEvents(this.branch, stamp);
    this.branch.client.stopSending = stopSending;
    return this
  });

  // GET
  define(leaf, 'get', function (path, val, stamp) {
    if (!stamp && val !== undefined) {
      stamp = createStamp(this.branch.stamp);
    }

    const subId = getApi(this.branch, this.id, path, val, stamp);
    if (subId) {
      if (stamp) {
        emitDataEvents(this.branch, stamp);
      }
      return new Leaf(this.branch, subId)
    }
  });

  // PARENT
  define(leaf, 'parent', function () {
    if (this.branch.leaves[this.id].parent) {
      return new Leaf(this.branch, this.branch.leaves[this.id].parent)
    }
  });

  // ROOT
  define(leaf, 'root', function () {
    return new Leaf(this.branch, root)
  });

  // ORIGIN
  define(leaf, 'origin', function () {
    return new Leaf(this.branch, origin(this.branch, this.id))
  });

  // COMPUTE
  define(leaf, 'compute', function () {
    return compute(this.branch, this.branch.leaves[this.id])
  });

  // PATH
  define(leaf, 'path', function () {
    return path(this.branch, this.id)
  });

  // INSPECT
  define(leaf, 'inspect', function () {
    return inspect(this.branch, this.branch.leaves[this.id])
  });

  // SERIALIZE
  define(leaf, 'serialize', function () {
    return serialize(this.branch, this.branch.leaves[this.id])
  });

  // FOREACH
  define(leaf, 'forEach', function (cb) {
    return forEach(this.branch, this.id, cb)
  });

  // MAP
  define(leaf, 'map', function (cb) {
    return map(this.branch, this.id, cb)
  });

  // FILTER
  define(leaf, 'filter', function (cb) {
    return filter(this.branch, this.id, cb)
  });

  // FIND
  define(leaf, 'find', function (cb) {
    return find(this.branch, this.id, cb)
  });

  // REDUCE
  define(leaf, 'reduce', function (cb, accumulator) {
    return reduce(this.branch, this.id, cb, accumulator)
  });

  // ON
  define(leaf, 'on', function (event, cb) {
    if (typeof event === 'function') {
      cb = event;
      event = 'data';
    }

    return on(this.branch, this.id, event, cb)
  });

  // SUBSCRIBE
  define(leaf, 'subscribe', function (options, cb) {
    if (typeof options === 'function') {
      cb = options;
      options = {};
    }

    return subscribe(this.branch, this.id, options, cb)
  });

  // EMIT
  define(leaf, 'emit', function (event, val, stamp) {
    if (!stamp) {
      stamp = createStamp(this.branch.stamp);
    }

    emit(this.branch, this.id, event, val, stamp);

    if (this.branch.client.queue && event !== 'data') {
      this.branch.client.queue.e.push([this.id, event, val, stamp]);
      drainQueue(this.branch.client);
    }

    return this
  });

  // LISTEN
  define(leaf, 'listen', function (port) {
    return listen(this.branch, port)
  });

  // CONNECT
  define(leaf, 'connect', function (url) {
    return connect(this.branch, url)
  });

  // SWITCH BRANCH
  define(leaf, 'switchBranch', function (branchKey) {
    if (this.branch.client.queue) {
      this.branch.client.queue.b = branchKey;
      sendAllSubscriptions(this.branch);
    }
  });
};

defineApi(Leaf.prototype);

exports.create = create;
exports.createPersist = createPersist;
exports.createStamp = createStamp;
